import fs from 'fs';
import initSqlJs from 'sql.js';
import SQL from 'sql.js';

const systemStrings = new Map<number, string>();
const setcodeStrings = new Map<number, string>();
const stringsMap = {
  system: systemStrings,
  setname: setcodeStrings,
} as const;

const attrOffset = 1010;
const raceOffset = 1020;
const typeOffset = 1050;

function getStringValueByMysticalNumber(offset: number, num: number) {
  const stringInfos: string[] = [];
  for (let i = 0; i < 32; i++) {
    if (num & (1 << i)) {
      const index = offset + i;
      const entry = systemStrings.get(index);
      if (entry) {
        stringInfos.push(entry);
      }
    }
  }
  return stringInfos;
}

function getSets(num: number) {
  const possibleSetValues = [
    num & 0xffff,
    (num >>> 16) & 0xffff,
    // (num >>> 32) & 0xffff,
    // (num >>> 48) & 0xffff,
  ].filter((x) => x !== 0);
  // console.error(possibleSetValues)

  return possibleSetValues.flatMap((set) => {
    const baseCode = set & 0xfff;
    if (!baseCode) return [];
    const baseSet = setcodeStrings.get(baseCode);
    if (!baseSet) return [];
    const extraCode = set & 0xf000;
    if (!extraCode) return [baseSet];
    // every bit combination of extra code is a possible extra set
    const possibleExtraCodes: number[] = [];
    for (let i = 1; i < 16; i++) {
      const extraCode = baseCode | (i << 12);
      if ((extraCode & set) === extraCode) {
        possibleExtraCodes.push(extraCode);
      }
    }
    // console.error(possibleExtraCodes);
    return [
      baseSet,
      ...possibleExtraCodes
        .map((extraCode) => {
          const extraSet = setcodeStrings.get(extraCode);
          if (!extraSet) return;
          return extraSet;
        })
        .filter((s) => s),
    ];
  });
}

async function loadString(path: string) {
  const stringsFile = await fs.promises.readFile(path, 'utf8');
  const stringsLines = stringsFile.split('\n');
  for (const line of stringsLines) {
    const [type, idStr, string] = line.split(' ');
    if (!type.startsWith('!')) continue;
    const map = stringsMap[type.slice(1)];
    if (!map) continue;
    const id = parseInt(idStr);
    map.set(id, string.split('\t')[0]);
  }
}

async function querySQL<T = any>(
  db: SQL.Database,
  sql: string,
  params: any = {},
) {
  const statement = db.prepare(sql);
  statement.bind(params);
  const results: T[] = [];
  while (statement.step()) {
    results.push(statement.getAsObject() as any);
  }
  statement.free();
  return results;
}

interface YGOProCardLike {
  id: number;
  name: string;
  desc: string;
  ot: number;
  alias: number;
  setcode: number;
  type: number;
  atk: number;
  def: number;
  level: number;
  race: number;
  attribute: number;
  category: number;
}

interface YGOProCard extends YGOProCardLike {
  displayLevel: number;
  cardLScale: number;
  cardRScale: number;
  displayAtk: string;
  displayDef: string;
  displayRace: string[];
  displayAttribute: string[];
  displayType: string[];
  linkMarkers: string[];
  sets: string[];
  overallString: string;
  picUrl: string;
}

const ygoproConstants = {
  TYPES: {
    TYPE_MONSTER: 1,
    TYPE_SPELL: 2,
    TYPE_TRAP: 4,
    TYPE_NORMAL: 16,
    TYPE_EFFECT: 32,
    TYPE_FUSION: 64,
    TYPE_RITUAL: 128,
    TYPE_TRAPMONSTER: 256,
    TYPE_SPIRIT: 512,
    TYPE_UNION: 1024,
    TYPE_DUAL: 2048,
    TYPE_TUNER: 4096,
    TYPE_SYNCHRO: 8192,
    TYPE_TOKEN: 16384,
    TYPE_QUICKPLAY: 65536,
    TYPE_CONTINUOUS: 131072,
    TYPE_EQUIP: 262144,
    TYPE_FIELD: 524288,
    TYPE_COUNTER: 1048576,
    TYPE_FLIP: 2097152,
    TYPE_TOON: 4194304,
    TYPE_XYZ: 8388608,
    TYPE_PENDULUM: 16777216,
    TYPE_SPSUMMON: 33554432,
    TYPE_LINK: 67108864,
  },
  LINK_MARKERS: {
    LINK_MARKER_BOTTOM_LEFT: 1,
    LINK_MARKER_BOTTOM: 2,
    LINK_MARKER_BOTTOM_RIGHT: 4,
    LINK_MARKER_LEFT: 8,
    LINK_MARKER_RIGHT: 32,
    LINK_MARKER_TOP_LEFT: 64,
    LINK_MARKER_TOP: 128,
    LINK_MARKER_TOP_RIGHT: 256,
  },
};

function isType(value: number, type: keyof typeof ygoproConstants.TYPES) {
  return !!(value & ygoproConstants.TYPES[type]);
}

function toDisplayAdValue(value: number) {
  if (value < 0) {
    return '?';
  }
  return value.toString();
}

function getMetaText(data: Partial<YGOProCard>) {
  const lines: string[] = [];
  const isMonster = isType(data.type, 'TYPE_MONSTER');
  let typeString = `[${data.displayType.join('|')}]`;
  if (isMonster) {
    typeString += ` ${data.displayRace[0]}/${data.displayAttribute[0]}`;
  }
  lines.push(typeString);
  if (isMonster) {
    const starString = isType(data.type, 'TYPE_XYZ')
      ? '\u2606'
      : isType(data.type, 'TYPE_LINK')
      ? 'LINK-'
      : '\u2605';
    let monsterStatus = `[${starString}${data.displayLevel}] ${data.displayAtk}/${data.displayDef}`;
    if (isType(data.type, 'TYPE_PENDULUM')) {
      monsterStatus += `  ${data.cardLScale}/${data.cardRScale}`;
    }
    if (isType(data.type, 'TYPE_LINK')) {
      monsterStatus += `  ${data.linkMarkers.join('')}`;
    }
    lines.push(monsterStatus);
  }

  return lines.join('  ');
}

function formatCard(data: YGOProCardLike): YGOProCard {
  const result: Partial<YGOProCard> = {
    ...data,
    displayLevel: data.level & 0xff,
    cardLScale: (data.level >>> 6) & 0xff,
    cardRScale: (data.level >>> 4) & 0xff,
    displayAtk: toDisplayAdValue(data.atk),
    displayRace: getStringValueByMysticalNumber(raceOffset, data.race),
    displayAttribute: getStringValueByMysticalNumber(
      attrOffset,
      data.attribute,
    ),
    displayType: getStringValueByMysticalNumber(typeOffset, data.type),
    sets: getSets(data.setcode),
  };
  if (!isType(data.type, 'TYPE_LINK')) {
    result.displayDef = toDisplayAdValue(data.def);
  } else {
    // result.name+="[LINK-" + cardLevel + "]";
    // result.name += " " + (result.atk < 0 ? "?" : result.atk) + "/- ";
    result.linkMarkers = [];
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_TOP_LEFT)
      result.linkMarkers.push('[↖]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_TOP)
      result.linkMarkers.push('[↑]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_TOP_RIGHT)
      result.linkMarkers.push('[↗]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_LEFT)
      result.linkMarkers.push('[←]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_RIGHT)
      result.linkMarkers.push('[→]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_BOTTOM_LEFT)
      result.linkMarkers.push('[↙]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_BOTTOM)
      result.linkMarkers.push('[↓]');
    if (result.def & ygoproConstants.LINK_MARKERS.LINK_MARKER_BOTTOM_RIGHT)
      result.linkMarkers.push('[↘]');
    result.displayDef = '-';
  }
  result.overallString = getMetaText(result);
  result.picUrl =
    (process.env.YGOPRO_PIC_URL_PREFIX ||
      'https://cdn02.moecube.com:444/images/ygopro-images-zh-CN/') +
    data.id.toString() +
    (process.env.YGOPRO_PIC_URL_SUFFIX || '.jpg');
  return result as YGOProCard;
}

async function main() {
  await Promise.all(process.argv.slice(3).map((path) => loadString(path)));
  const SQL = await initSqlJs();

  const dbBuffer = await fs.promises.readFile(process.argv[2]);
  const db = new SQL.Database(dbBuffer);
  const cards = await querySQL<YGOProCardLike>(
    db,
    'select datas.*,texts.name,texts.desc from datas,texts where datas.id = texts.id and datas.type & 0x4000 = 0 and (datas.alias = 0 or datas.id - datas.alias > 10 or datas.id - datas.alias < -10)',
  );
  const formattedCards = cards.map(formatCard);
  console.log(JSON.stringify(formattedCards, null, 2));
  process.exit(0);
}
main();
